(************************************************************************)
(*  v      * Internal Extraction                                        *)
(* <O___,, *                                                            *)
(*   \VV/  **************************************************************)
(*    //   * © 2007-2009 Stéphane Glondu                                *)
(*         * File distributed under the terms of the CeCILL 2.0 license *)
(************************************************************************)

(* This file needs "Internal Extraction" command! *)

Require Import List.
Require Import Ml MlSemExtra.
Require Import InternalExtraction.

Set Implicit Arguments.
Unset Strict Implicit.

(************************************************************************)

(** * nat are extractible *)

(*
Inductive nat : Set :=  O : nat | S : nat -> nat

For S: Argument scope is [nat_scope]
*)

Inductive nat_pred : nat -> term -> Prop :=
  | nat_predO : nat_pred O (TConstr 0 nil)
  | nat_predS : forall n n', nat_pred n n' -> nat_pred (S n) (TConstr 1 (n'::nil)).

Lemma nat_value : forall n n', nat_pred n n' -> IsValue n'.
Proof.
  induction 1; auto.
Qed.

Lemma nat_closed : forall n n', nat_pred n n' -> clos_after 0 n'.
Proof.
  induction 1; auto.
Qed.

Instance nat_extractible : InternallyExtractible nat := {
  ie_pred := nat_pred;
  ie_value := nat_value;
  ie_closed := nat_closed
}.

Lemma O__simul : simul 0 (TConstr 0 nil).
Proof.
  do 2 esplit; [ reflexivity | constructor | ].
  repeat constructor; auto; fail.
Qed.

Lemma S__simul : forall n n', simul n n' -> simul (S n) (TConstr 1 (n'::nil)).
Proof.
  intros n n' Hnn'.
  destruct Hnn' as [ n'' [ ? ? Hclos ] Hnn'' ].
  exists (TConstr 1 (n''::nil)); constructor.
  - eauto using SmallSteps_constr; fail.
  - repeat constructor; assumption.
  - repeat constructor; assumption.
Qed.

Hint Resolve O__simul S__simul : internal_extraction.

Ltac ie_change_nat :=
  repeat lazymatch goal with
    | H : nat_pred ?a ?b |- _ => change (ie_pred a b) in H
  end.

Ltac ie_destruct_nat Hnn' :=
  lazymatch goal with
    |- ?A _ _ =>
      refine (match Hnn' in nat_pred x x' return A x x' with
                | nat_predO => _
                | nat_predS a b Hie => _
              end)
  end; [ | ie_change_nat ].

(************************************************************************)

(** * Eta-expanded constructor *)

Definition succ := S.
Internal Extraction succ.

Lemma succ__simul : simul succ succ__extr.
Proof.
  unfold succ, succ__extr.
  ie_hnf.
  intros; ie_simpl; ie_easy.
Qed.

Hint Resolve succ__simul : internal_extraction.

(************************************************************************)

(** * A trivial fixpoint *)

Fixpoint test n :=
  match n with
    | O => O
    | S n => test n
  end.

Internal Extraction test.

Lemma test__simul : simul test test__extr.
Proof.
  unfold test, test__extr.
  ie_hnf.
  fix IHn 3.
  intros n n' Hxx'.
  ie_simpl.

  pattern n at 1, n' at 1.
  ie_destruct_nat Hxx'; ie_simpl; ie_easy.
Qed.

Hint Resolve test__simul : internal_extraction.

(************************************************************************)

(** * Ackermann function *)

Fixpoint ack n :=
  match n with
    | O => S
    | S n =>
      (fix F (m:nat) :=
        match m with
          | O => ack n 1
          | S m => ack n (F m)
        end)
  end.

Internal Extraction ack.

Lemma ack__simul : simul ack ack__extr.
Proof.
  unfold ack, ack__extr.
  ie_hnf.
  fix IHn 3; intros n n' Hnn'.
  ie_simpl.
  pattern n at 1, n' at 1.
  ie_destruct_nat Hnn'; ie_simpl.

  (* O *)
  ie_hnf; intros; ie_simpl; ie_easy.

  (* S *)
  ie_hnf; fix IHm 3.
  intros m m' Hmm'; ie_simpl.
  pattern m at 1, m' at 1.
  ie_destruct_nat Hmm'; ie_simpl; ie_apply; ie_easy.
Qed.

Hint Resolve ack__simul : internal_extraction.

(************************************************************************)

(** * Basic arithmetics *)

Internal Extraction pred.

Lemma pred__simul : simul pred pred__extr.
Proof.
  unfold pred, pred__extr.
  ie_hnf.
  intros n n' Hnn'.
  ie_simpl.

  pattern n at 1, n' at 1.
  ie_destruct_nat Hnn'; ie_simpl; ie_easy.
Qed.

Hint Resolve pred__simul : internal_extraction.

Fixpoint test1 n m {struct n} :=
  match n with
    | O => O
    | S p => test1 p (match p with 0 => 0 | S q => test1 p q end)
  end.

Internal Extraction test1.

Lemma test1__simul : simul test1 test1__extr.
Proof.
  unfold test1, test1__extr.
  ie_hnf.
  fix IHn 3.

  intros n n' Hxx'.
  ie_simpl.
  ie_hnf.
  intros m m' Hmm'.
  ie_simpl.

  pattern n at 1, n' at 1.
  ie_destruct_nat Hxx'; ie_simpl.

  (* O *)
  ie_easy.

  (* S *)
  ie_apply.

  (* this is the induction hypothesis... *)
  ie_easy.

  (* here is the nested pattern matching... *)
  pattern a at 1, b at 1.
  ie_destruct_nat Hie; ie_simpl.

  (* S O *)
  ie_easy.

  (* S S q0 *)
  ie_apply; ie_easy.
Qed.

Hint Resolve test1__simul : internal_extraction.


Internal Extraction plus.


(*
plus =
fix plus (n m : nat) : nat := match n with
                              | 0 => m
                              | S p => S (plus p m)
                              end
     : nat -> nat -> nat

Argument scopes are [nat_scope nat_scope]

plus__extr =
TFix
  (TFun
     (TMatch (TVar 1)
        (Patc 0 (TVar 0)
         :: Patc 1 (TConstr 1 (TVar 3 @ TVar 0 @ TVar 1 :: nil)) :: nil)))
     : term
*)

Lemma plus__simul : simul plus plus__extr.
Proof.
  unfold plus, plus__extr.
  ie_hnf.
  fix IHn 3.
  intros n n' Hnn'.
  ie_simpl.
  ie_hnf.
  intros m m' Hmm'.
  ie_simpl.
  pattern n at 1, n' at 1.
  ie_destruct_nat Hnn'; ie_simpl.

  (* O *)
  ie_easy.

  (* S *)
  apply S__simul.
  ie_apply; ie_easy.
Qed.

Hint Resolve plus__simul : internal_extraction.


Internal Extraction mult.

Lemma mult__simul : simul mult mult__extr.
Proof.
  unfold mult, mult__extr.
  destruct plus__simul.
  ie_hnf.
  fix IHn 3.

  intros n n' Hxx'.
  ie_simpl.
  ie_hnf.
  intros m m' Hmm'.
  ie_simpl.

  pattern n at 1, n' at 1.
  ie_destruct_nat Hxx'; ie_simpl.

  (* O *)
  ie_easy.

  (* S *)
  ie_apply; ie_apply; ie_easy.
Qed.

Hint Resolve mult__simul : internal_extraction.


Definition pred2 : forall n, n <> 0 -> nat.
Proof.
  intros n H.
  destruct n; intros.
  destruct H; auto.
  exact n.
Defined.

Internal Extraction pred2.

Lemma pred2__simul : simul pred2 pred2__extr.
Proof.
  unfold pred2, pred2__extr.
  ie_hnf.
  intros n n' Hxx'.
  ie_simpl.
  ie_hnf.

  change ((fun N N' =>
   forall (x : N <> 0) (x' : term),
   ie_pred x x' ->
   simul (match N as n0 return (n0 <> 0 -> nat) with
   | 0 => fun H : 0 <> 0 => match H (eq_refl 0) return nat with
                            end
   | S n0 => fun _ : S n0 <> 0 => n0
   end x)
   (TFun
        (TMatch N'
           (Patc 0 (TFun (TMatch TDummy nil))
            :: Patc 1 (TFun (TVar 1)) :: nil) @ TDummy) @ x')
  ) n n').

  ie_destruct_nat Hxx'; intros h h' Hhh'; ie_simpl.

  (* the perfect subgoal! *)
  destruct h.

  (* trivialities *)
  ie_easy.
Qed.

Hint Resolve pred2__simul : internal_extraction.

(************************************************************************)

(** * Polymorphism: an open type, list *)

Section list_pred.

  Variable A : Type.
  Hypothesis IEA : InternallyExtractible A.

  (*
    Inductive list (A : Type) : Type :=
      nil : list A | cons : A -> list A -> list A
  *)

  Inductive list_pred : list A -> term -> Prop :=
    | nil_pred : list_pred nil (TConstr 0 nil)
    | cons_pred :
      forall x x', ie_pred x x' ->
        forall xs xs', list_pred xs xs' ->
          list_pred (x::xs) (TConstr 1 (x'::xs'::nil)).

  Lemma list_value : forall X X', list_pred X X' -> IsValue X'.
  Proof.
    fix 3; destruct 1; repeat constructor.
    - apply ie_value with x; assumption.
    - apply list_value with xs; assumption.
  Qed.

  Lemma list_closed : forall X X', list_pred X X' -> clos_after 0 X'.
  Proof.
    fix 3; destruct 1; repeat constructor.
    - apply ie_closed with x; assumption.
    - apply list_closed with xs; assumption.
  Qed.

  Instance list_extractible : InternallyExtractible (list A) := {
    ie_pred := list_pred;
    ie_value := list_value;
    ie_closed := list_closed
  }.

  Lemma nil__simul : simul (@nil A) (TConstr 0 nil).
  Proof.
    exists (TConstr 0 nil); split.
    - eauto using SmallSteps_constr; fail.
    - repeat constructor; assumption.
    - repeat constructor; assumption.
  Qed.

  Lemma cons__simul :
    forall (x : A) x', simul x x' ->
      forall xs xs', simul xs xs' ->
        simul (x::xs) (TConstr 1 (x'::xs'::nil)).
  Proof.
    intros x x' Hxx'.
    destruct Hxx' as [ x'' [ ? ? Hclosx ] Hxx'' ].
    intros xs xs' Hxsxs'.
    destruct Hxsxs' as [ xs'' [ ? ? Hclosxs ] Hxsxs'' ].
    exists (TConstr 1 (x''::xs''::nil)); split.
    - eauto using SmallSteps_constr; fail.
    - repeat constructor; assumption.
    - repeat constructor; assumption.
  Qed.

End list_pred.

Existing Instance list_extractible.
Hint Resolve nil__simul cons__simul : internal_extraction.

Ltac ie_change_list :=
  repeat lazymatch goal with
    | H : list_pred _ ?a ?b |- _ => change (ie_pred a b) in H
  end.

Ltac ie_destruct_list HXX' :=
  lazymatch goal with
    |- ?A _ _ =>
      refine (match HXX' in @list_pred _ _ x x' return A x x' with
                | nil_pred => _
                | cons_pred x x' Hxx' xs xs' Hxsxs' => _
              end)
  end; [ | ie_change_list ].

Internal Extraction length.

Section length__simul.

  Variables A : Type.
  Hypothesis IEA : InternallyExtractible A.

  Lemma length__simul : simul (@length A) (length__extr @ TDummy).
  Proof.
    unfold length, length__extr.
    ie_hnf.
    fix 3.
    intros l l' Hll'.
    ie_simpl.

    pattern l at 1, l' at 1.
    ie_destruct_list Hll'; ie_simpl; ie_easy.
  Qed.

End length__simul.

Hint Resolve length__simul : internal_extraction.

Internal Extraction map.

Section map__simul.

  Variables A : Type.
  Hypothesis IEA : InternallyExtractible A.

  Variables B : Type.
  Hypothesis IEB : InternallyExtractible B.

  Lemma map__simul : simul (@map A B) (map__extr @ TDummy @ TDummy).
  Proof.
    unfold map, map__extr.
    ie_hnf.
    intros f f' Hff'.
    ie_hnf.
    fix 3.
    intros l l' Hll'.
    ie_simpl.

    pattern l at 1, l' at 1.
    ie_destruct_list Hll'; ie_simpl.
    ie_easy.
    apply cons__simul.
    ie_apply; ie_easy.
    ie_easy.
  Qed.

End map__simul.

Hint Resolve map__simul : internal_extraction.

(************************************************************************)

(** * Higher-order *)

(* Another kind of polymorphism *)

Definition apply A B (f : forall x:A, B x) x := f x.
Internal Extraction apply.

Lemma apply__simul :
  forall A (IEA: InternallyExtractible A),
    forall B (IEB: forall x, InternallyExtractible (B x)),
      simul (@apply A B) (apply__extr @ TDummy @ TDummy).
Proof.
  unfold apply, apply__extr.
  intros; ie_hnf; intros; ie_simpl; ie_hnf.
  intros; ie_simpl; ie_apply; ie_easy.
Qed.

Hint Resolve apply__simul : internal_extraction.

(************************************************************************)

(** * A datatype with logical subterms, and no free variable *)

Inductive even :=
  even_intro : forall n, (exists p, n = p + p) -> even.

Inductive even_pred : even -> term -> Prop :=
  even_pred_intro :
    forall n n', ie_pred n n' ->
      forall h h', ie_pred h h' ->
        even_pred (@even_intro n h) (TConstr 0 (n'::h'::nil)).

Lemma even_value : forall n n', even_pred n n' -> IsValue n'.
Proof.
  destruct 1; repeat constructor.
  - apply ie_value with n; assumption.
  - apply ie_value with h; assumption.
Qed.

Lemma even_closed : forall n n', even_pred n n' -> clos_after 0 n'.
Proof.
  destruct 1; repeat constructor.
  - apply ie_closed with n; assumption.
  - apply ie_closed with h; assumption.
Qed.

Instance even_extractible : InternallyExtractible even := {
  ie_pred := even_pred;
  ie_value := even_value;
  ie_closed := even_closed
}.

Lemma even_intro__simul :
  forall n n', simul n n' ->
    forall h h', simul h h' ->
      simul (@even_intro n h) (TConstr 0 (n'::h'::nil)).
Proof.
  intros n n' Hnn'.
  destruct Hnn' as [ n'' [ ? ? Hclosn ] Hnn'' ].
  intros h h' Hhh'.
  destruct Hhh' as [ h'' [ ? ? Hclosh ] Hhh'' ].
  exists (TConstr 0 (n''::h''::nil)); split.
  - eauto using SmallSteps_constr; fail.
  - repeat constructor; assumption.
  - repeat constructor; assumption.
Qed.

Hint Resolve even_intro__simul : internal_extraction.

Ltac ie_change_even :=
  repeat lazymatch goal with
    | H : even_pred ?a ?b |- _ => change (even_pred a b) in H
  end.

Ltac ie_destruct_even Hnn' :=
  lazymatch goal with
    |- ?A _ _ =>
      refine (match Hnn' in even_pred x x' return A x x' with
                | even_pred_intro n n' Hnn' h h' Hhh' => _
              end)
  end; [ | ie_change_even ].

Lemma ex_intro__simul :
  forall A (P : A -> Prop) (x : A) (H : P x),
    simul (ex_intro P x H) TDummy.
Proof.
  intros; exists TDummy; split.
  - reflexivity.
  - reflexivity.
  - constructor.
Qed.

Hint Resolve ex_intro__simul : internal_extraction.

Definition double : nat -> even.
Proof.
  refine (fun p => @even_intro (p + p) _).
  exists p; reflexivity.
Defined.

Internal Extraction double.

Lemma double__simul : simul double double__extr.
Proof.
  unfold double, double__extr.
  destruct plus__simul.
  ie_hnf.

  intros n n' Hnn'.
  ie_simpl.
  apply even_intro__simul.
  ie_apply; [ ie_apply | ]; ie_easy.
  apply ex_intro__simul.
Qed.

Hint Resolve double__simul : internal_extraction.

(************************************************************************)

(** * Another datatype with logical subterms *)

(* This one has a free variable, which is not supported... Let's try
   anyway, and see how far we can go... *)

Section sig_pred.

  Variable n : nat.

  Inductive predT :=
    predT_intro : forall p, n = S p -> predT.

  Inductive predT_pred : predT -> term -> Prop :=
    predT_pred_intro :
      forall p p', ie_pred p p' ->
        forall h h', ie_pred h h' ->
          predT_pred (@predT_intro p h) (TConstr 0 (p'::h'::nil)).

  Lemma predT_value : forall X X', predT_pred X X' -> IsValue X'.
  Proof.
    destruct 1; repeat constructor.
    - apply ie_value with p; assumption.
    - apply ie_value with h; assumption.
  Qed.

  Lemma predT_closed : forall X X', predT_pred X X' -> clos_after 0 X'.
  Proof.
    destruct 1; repeat constructor.
    - apply ie_closed with p; assumption.
    - apply ie_closed with h; assumption.
  Qed.

  Instance predT_extractible : InternallyExtractible predT := {
    ie_pred := predT_pred;
    ie_value := predT_value;
    ie_closed := predT_closed
  }.

  Lemma predT_intro__simul :
    forall p p', simul p p' ->
      forall h h', simul h h' ->
        simul (@predT_intro p h) (TConstr 0 (p'::h'::nil)).
  Proof.
    intros p p' Hpp'.
    destruct Hpp' as [ p'' [ ? ? Hclosp ] Hpp'' ].
    intros h h' Hhh'.
    destruct Hhh' as [ h'' [ ? ? Hclosh ] Hhh'' ].
    exists (TConstr 0 (p''::h''::nil)); split.
    - eauto using SmallSteps_constr; fail.
    - repeat constructor; assumption.
    - repeat constructor; assumption.
  Qed.

End sig_pred.

Existing Instance predT_extractible.
Hint Resolve predT_intro__simul : internal_extraction.

Ltac ie_change_predT :=
  repeat lazymatch goal with
    | H : predT_pred _ ?a ?b |- _ => change (ie_pred a b) in H
  end.

Ltac ie_destruct_predT HXX' :=
  lazymatch goal with
    |- ?A _ _ =>
      refine (match HXX' in @predT_pred _ x x' return A x x' with
                | predT_pred_intro p p' Hpp' h h' Hhh' => _
              end)
  end; [ | ie_change_predT ].

Definition pred3 : forall n, n <> 0 -> predT n.
Proof.
  intros; destruct n.
  destruct H; simpl; auto.
  exists n; auto.
Defined.

Internal Extraction pred3.

Lemma pred3__simul : simul pred3 pred3__extr.
Proof.
  unfold pred3, pred3__extr.
  ie_hnf.
  intros n n' Hnn'.
  ie_simpl.

  (* it's too hard with dependencies :-( *)
  Fail pattern n at 1, n' at 1.

  (* but we can manage manually *)
  change (
   (fun n n' => simul
     (fun H : n <> 0 =>
      match n as n0 return (n0 <> 0 -> predT n0) with
      | 0 => fun H0 : 0 <> 0 => match H0 eq_refl return (predT 0) with
                                end
      | S n0 => fun _ : S n0 <> 0 => predT_intro eq_refl
      end H)
     (TFun
        (TMatch n'
           (Patc 0 (TFun (TMatch TDummy nil))
            :: Patc 1 (TFun (TConstr 0 (TVar 1 :: TDummy :: nil))) :: nil) @
         TDummy))
   ) n n'
  ).
  ie_destruct_nat Hnn'; ie_simpl.

  (* O *)
  ie_hnf; intros h h' Hhh'; ie_simpl.
  destruct (h eq_refl).

  (* S *)
  ie_hnf; intros h h' Hhh'; ie_simpl.
  apply predT_intro__simul.
  ie_easy.
  exists TDummy; split.
  - reflexivity.
  - reflexivity.
  - constructor.
Qed.

Hint Resolve pred3__simul : internal_extraction.

(************************************************************************)
