(************************************************************************)
(*  v      * Internal Extraction                                        *)
(* <O___,, *                                                            *)
(*   \VV/  **************************************************************)
(*    //   * © 2007-2009 Stéphane Glondu                                *)
(*         * File distributed under the terms of the CeCILL 2.0 license *)
(************************************************************************)

Require Import Arith Bool List Setoid.
Require Import Ml MlSemExtra.

(* For convenience, not really used in this file *)
Declare ML Module "intextr".

Set Implicit Arguments.
Unset Strict Implicit.

Local Generalizable Variables A B.

(************************************************************************)

(** * Basic definitions *)

Class InternallyExtractible A := {
  ie_pred : A -> term -> Prop;
  ie_value : forall x x', ie_pred x x' -> IsValue x';
  ie_closed : forall x x', ie_pred x x' -> clos_after 0 x'
}.

Record extracted `{IEA : InternallyExtractible A} x x' x'' := {
  ie_ered : x' ===> x'';
  ie_epred : ie_pred x x'';
  ie_eclosed : clos_after 0 x'
}.

Definition simul `{IEA : InternallyExtractible A} x x' :=
  exists x'', extracted x x' x''.

Lemma ie_from `{IEA : InternallyExtractible A} :
  forall x x', simul x x' -> forall x'', (x'' ===> x') -> clos_after 0 x'' -> simul x x''.
Proof.
  intros.
  destruct H as [ x2 [ Hredx2 Hxx2 Hclos ] ].
  do 2 esplit; [ eapply SmallSteps_trans2 | | ]; eauto.
Qed.

Ltac fromtac :=
  match goal with
    |- simul ?x ?x'' =>
      refine (
        (fun (Hxx': simul x _) =>
          let Hred := (_ : x'' ===> _) in
          let Hclos := (_ : clos_after 0 x'') in
          ie_from Hxx' Hred Hclos)
        _)
  end
.

Lemma ie_compat `{IEA : InternallyExtractible A} :
  forall (x:A) x', ie_pred x x' -> simul x x'.
Proof.
  intros; do 2 esplit;
    [ reflexivity
    | assumption
    | eapply ie_closed; eassumption
    ].
Qed.

(************************************************************************)

(** * Products are extractible *)

Section prod_extractible.

  Variable A : Type.
  Hypothesis IEA : InternallyExtractible A.

  Variable B : A -> Type.
  Hypothesis IEB : forall x, InternallyExtractible (B x).

  Record prod_pred (f : forall x, B x) f' := {
    ie_prod_value : IsValue f';
    ie_prod_clos : clos_after 0 f';
    ie_prod_pred : forall x x', ie_pred x x' -> simul (f x) (f' @ x')
  }.

  Lemma prod_value : forall f f', prod_pred f f' -> IsValue f'.
  Proof.
    intros f f' [ [ ] ]; auto.
  Qed.

  Lemma prod_closed : forall f f', prod_pred f f' -> clos_after 0 f'.
  Proof.
    intros f f' [ [ ] ]; auto.
  Qed.

  Instance prod_extractible : InternallyExtractible (forall x, B x) := {
    ie_pred := prod_pred;
    ie_value := prod_value;
    ie_closed := prod_closed
  }.

  Lemma ie_compat_app :
    forall (f : forall x, B x) f', simul f f' ->
      forall x x', simul x x' -> simul (f x) (f' @ x').
  Proof.
    intros f f' [f'' [Hrf [Hvf Hcf'' Hpf] Hcf']] x x' [x'' [Hrx Hpx Hcx]].
    fromtac.
    eapply SmallSteps_trans2; [eapply SmallSteps_app1 | eapply SmallSteps_app2]; eassumption.
    constructor; assumption.
    auto; fail.
  Qed.

End prod_extractible.

Existing Instance prod_extractible.

(************************************************************************)

(** * Erased stuff are extractible *)

Definition is_TDummy_pred A (x:A) x' := x' = TDummy.

Lemma is_TDummy_value : forall A (x:A) x', is_TDummy_pred x x' -> IsValue x'.
Proof.
  inversion 1; constructor.
Qed.

Lemma is_TDummy_closed : forall A (x:A) x', is_TDummy_pred x x' -> clos_after 0 x'.
Proof.
  inversion 1; constructor.
Qed.

Instance pred_extractible (A:Prop) : InternallyExtractible A := {
  ie_pred := @is_TDummy_pred A;
  ie_value := @is_TDummy_value A;
  ie_closed := @is_TDummy_closed A
}.

Inductive sort_pred : Type -> term -> Prop :=
  sort_pred_intro : forall x, InternallyExtractible x -> sort_pred x TDummy.

Lemma sort_value : forall x x', sort_pred x x' -> IsValue x'.
Proof.
  destruct 1; constructor.
Qed.

Lemma sort_closed : forall x x', sort_pred x x' -> clos_after 0 x'.
Proof.
  destruct 1; constructor.
Qed.

Instance sort_extractible : InternallyExtractible Type := {
  ie_pred := @is_TDummy_pred Type;
  ie_value := @is_TDummy_value Type;
  ie_closed := @is_TDummy_closed Type
}.

(************************************************************************)

(** * Hint database *)

Create HintDb internal_extraction discriminated.

Hint Resolve @ie_compat : internal_extraction.

Ltac ie_easy := auto with internal_extraction; fail.

(************************************************************************)

(** * Alternative definition of terms for easy manipulation in tactics *)

Inductive IE_term :=
  | IE_TDummy : IE_term
  | IE_TVar : nat -> IE_term
  | IE_TLet : IE_term -> IE_term -> IE_term
  | IE_TFun : IE_term -> IE_term
  | IE_TFix : IE_term -> IE_term
  | IE_TApply : IE_term -> IE_term -> IE_term
  | IE_TConstr : nat -> list IE_term -> IE_term
  | IE_TMatch : IE_term -> list (nat * IE_term) -> IE_term
  | IE_Value : forall `{IEA : InternallyExtractible A} x x', ie_pred x x' -> IE_term
  | IE_Extracted : forall `{IEA : InternallyExtractible A} x x' x'', extracted x x' x'' -> IE_term
.

Fixpoint IE_is_closed_after n t :=
  match t with
    | IE_TDummy => true
    | IE_TVar i => if le_lt_dec n i then false else true
    | IE_TLet a b => IE_is_closed_after n a && IE_is_closed_after (S n) b
    | IE_TFun f => IE_is_closed_after (S n) f
    | IE_TFix f => IE_is_closed_after (S (S n)) f
    | IE_TApply a b => IE_is_closed_after n a && IE_is_closed_after n b
    | IE_TConstr k tl => forallb (IE_is_closed_after n) tl
    | IE_TMatch t pl => IE_is_closed_after n t &&
      forallb (fun p:(nat*IE_term) => let (k,t) := p in IE_is_closed_after (k+n) t) pl
    | IE_Value _ _ _ _ _ => true
    | IE_Extracted _ _ _ _ _ _ => true
  end
.

Fixpoint IE_is_value t :=
  match t with
    | IE_TDummy => true
    | IE_TFun f => IE_is_closed_after 1 f
    | IE_TFix f => IE_is_closed_after 2 f
    | IE_TConstr k tl => forallb IE_is_value tl
    | IE_Value _ _ _ _ _ => true
    | _ =>
      (* NOTE: an IE_Extracted can be a value but, here, we care only
         about correction when return value is true *)
      false
  end
.

Fixpoint IE_subst n u t :=
  match t with
    | IE_TVar k =>
      match lt_eq_lt_dec k n with
        | inleft (left _) => IE_TVar k
        | inleft (right _) => u
        | inright _ => IE_TVar (pred k)
      end
    | IE_TFun t => IE_TFun (IE_subst (S n) u t)
    | IE_TFix t => IE_TFix (IE_subst (S (S n)) u t)
    | IE_TLet t1 t2 => IE_TLet (IE_subst n u t1) (IE_subst (S n) u t2)
    | IE_TApply t1 t2 => IE_TApply (IE_subst n u t1) (IE_subst n u t2)
    | IE_TConstr k tl => IE_TConstr k (map (IE_subst n u) tl)
    | IE_TMatch t pl =>
        IE_TMatch (IE_subst n u t)
          (map (fun p:(nat*IE_term) => let (k,t):=p in (k, (IE_subst (k+n) u t))) pl)
    | x => x
  end
.

Fixpoint IE_subst_list_iter n l t :=
  match l with
    | nil => t
    | u::l => IE_subst_list_iter n l (IE_subst n u t)
  end.

(************************************************************************)

(** * Going from term to IE_term and back again *)

Ltac IE_quote t cont :=
  let rec expand_patlist pl cont :=
    lazymatch pl with
      | nil => cont (@nil (nat * IE_term))
      | (Patc ?k ?t)::?pl =>
        IE_quote t ltac:(fun t =>
        expand_patlist pl ltac:(fun pl =>
        cont ((k,t)::pl)))
    end
  in
  lazymatch t with
    | TDummy => cont (IE_TDummy)
    | TVar ?k => cont (IE_TVar k)
    | TLet ?a ?b => IE_quote a ltac:(fun a => IE_quote b ltac:(fun b => cont (IE_TLet a b)))
    | TFun ?f => IE_quote f ltac:(fun f => cont (IE_TFun f))
    | TFix ?f => IE_quote f ltac:(fun f => cont (IE_TFix f))
    | TApply ?a ?b => IE_quote a ltac:(fun a => IE_quote b ltac:(fun b => cont (IE_TApply a b)))
    | TConstr ?k ?tl => IE_quote_list tl ltac:(fun tl => cont (IE_TConstr k tl))
    | TMatch ?t ?pl => IE_quote t ltac:(fun t => expand_patlist pl ltac:(fun pl => cont (IE_TMatch t pl)))
    | ?t =>
      lazymatch goal with
        | H : ie_pred _ t |- _ => cont (IE_Value H)
        | H : extracted _ _ t |- _ => cont (IE_Value (ie_epred H))
        | H : extracted _ t _ |- _ => cont (IE_Extracted H)
        | _ => fail "Internal Extraction: IE_quote failure"
      end
  end
with IE_quote_list tl cont :=
  lazymatch tl with
    | nil => cont (@nil IE_term)
    | ?t::?tl => IE_quote t ltac:(fun t => IE_quote_list tl ltac:(fun tl => cont (t::tl)))
  end.

Fixpoint IE_interp t :=
  match t with
    | IE_TDummy => TDummy
    | IE_TVar k => TVar k
    | IE_TLet a b => TLet (IE_interp a) (IE_interp b)
    | IE_TFun f => TFun (IE_interp f)
    | IE_TFix f => TFix (IE_interp f)
    | IE_TApply a b => TApply (IE_interp a) (IE_interp b)
    | IE_TConstr k tl => TConstr k (map IE_interp tl)
    | IE_TMatch t pl =>
      TMatch (IE_interp t)
        (map
          (fun p:(nat*IE_term) => let (k,t):=p in Patc k (IE_interp t))
          pl)
    | IE_Value _ _ _ t _ => t
    | IE_Extracted _ _ _ t _ _ => t
  end
.

(************************************************************************)

(** * Basic compatibility lemmas *)

Lemma IE_is_closed_after_correction : forall n t,
  IE_is_closed_after n t = true -> clos_after n (IE_interp t).
Proof.
  fix IHt 2.
  intros p t H; destruct t; simpl in *.
  constructor; fail.
  destruct (le_lt_dec p n); [ discriminate | constructor; assumption ].
  repeat rewrite andb_true_iff in H; constructor; apply IHt; intuition fail.
  repeat rewrite andb_true_iff in H; constructor; apply IHt; intuition fail.
  repeat rewrite andb_true_iff in H; constructor; apply IHt; intuition fail.
  repeat rewrite andb_true_iff in H; constructor; apply IHt; intuition fail.
  { (* constructors *)
    constructor; induction l; simpl in *; constructor;
      rewrite andb_true_iff in H.
    - apply IHt; intuition fail.
    - apply IHl; intuition fail.
  }
  { (* patterns *)
    repeat rewrite andb_true_iff in H; constructor.
    apply IHt; intuition fail.
    apply proj2 in H.
    induction l as [ | [k u] l]; simpl in *.
    - constructor.
    - repeat rewrite andb_true_iff in H; destruct H.
      { constructor.
        + constructor; apply IHt; assumption.
        + apply IHl; assumption.
      }
  }
  { (* values *)
    assert (Hc : clos_after 0 x') by (eapply ie_closed; eassumption).
    rewrite <- clos_after'_alt, <- clos_after_alt in Hc |- *.
    hnf in Hc |- *.
    intros; auto with arith.
  }
  { (* extracted *)
    assert (Hc : clos_after 0 x') by (eapply ie_eclosed; eassumption).
    rewrite <- clos_after'_alt, <- clos_after_alt in Hc |- *.
    hnf in Hc |- *.
    intros; auto with arith.
  }
Qed.

Lemma IE_is_value_correction : forall t, IE_is_value t = true -> IsValue (IE_interp t).
Proof.
  fix IHt 1; intros t H; destruct t; simpl in *; try constructor; try discriminate.
  induction l; simpl forallb in *.
    constructor.
    simpl; rewrite andb_true_iff in H; destruct H as (Ha, Hl); intuition.
  eapply ie_value; eassumption.
Qed.

Lemma IE_subst_correction : forall t n u,
  IE_interp (IE_subst n u t) = (IE_interp t)[n := IE_interp u].
Proof.
  fix IHt 1.
  intros t k u; destruct t; simpl; auto.
  destruct (lt_eq_lt_dec n k) as [ [ | ] | ]; intros; reflexivity.
  f_equal; induction l; simpl; auto; fail.
  f_equal; auto.
    induction l; simpl; auto.
    destruct a; auto; fail.
  symmetry; apply subst_clos; rewrite <- clos_alt; eapply ie_closed; eassumption.
  symmetry; apply subst_clos; rewrite <- clos_alt; eapply ie_eclosed; eassumption.
Qed.

Lemma IE_subst_list_iter_correction : forall l t n,
  IE_interp (IE_subst_list_iter n l t) = (IE_interp t)[n ;;= map IE_interp l].
Proof.
  induction l; intros t n; simpl; auto.
  rewrite <- IE_subst_correction.
  rewrite <- IHl.
  auto.
Qed.

(* This is bonus with IE_term (not true with regular terms) *)

Lemma IE_value_is_closed_bool :
  forall t, IE_is_value t = true -> IE_is_closed_after 0 t = true.
Proof.
  fix IHt 1; intros t H.
  destruct t; simpl in *; intuition.
  induction l; simpl in *.
  - reflexivity.
  - rewrite andb_true_iff in H |- *; intuition.
Qed.

Lemma IE_value_is_closed :
  forall t, IE_is_value t = true -> clos_after 0 (IE_interp t).
Proof.
  intros.
  apply IE_is_closed_after_correction.
  apply IE_value_is_closed_bool.
  assumption.
Qed.

(************************************************************************)

(** * Reduction *)

Fixpoint IE_red n t :=
  match n with
    | O => t
    | S n =>
      match t with
        | IE_TApply a b =>
          let b := IE_red n b in
          if IE_is_value b then
            let a := IE_red n a in
            if IE_is_value a then
              match a with
                | IE_TFun body => IE_red n (IE_subst O b body)
                | IE_TFix body => IE_red n (IE_subst O a (IE_subst O b body))
                | _ => IE_TApply a b
              end
            else IE_TApply a b
          else IE_TApply a b
        | IE_TMatch t pl =>
          let t := IE_red n t in
          if IE_is_value t then
            match t with
              | IE_TConstr k tl =>
                match nth_error pl k with
                  | Some (pn, pt) =>
                    if eq_nat_dec pn (length tl) then
                      IE_subst_list_iter O (rev tl) pt
                    else
                      IE_TMatch t pl
                  | _ => IE_TMatch t pl
                end
              | _ => IE_TMatch t pl
            end
          else IE_TMatch t pl
        | IE_Extracted _ _ _ _ _ H => IE_Value (ie_epred H)
        | _ => t
      end
  end.

Lemma IE_red_correction : forall n t, IE_interp t ===> IE_interp (IE_red n t).
Proof.
  induction n; intros t; simpl.
  reflexivity.
  destruct t; try reflexivity.
  case_eq (IE_is_value (IE_red n t2)); intros Heq2;
    [|simpl; auto using SmallSteps_app2; fail].
  case_eq (IE_is_value (IE_red n t1)); intros Heq1;
    [|simpl; eauto using SmallSteps_trans2, SmallSteps_app1, SmallSteps_app2; fail].
  generalize (refl_equal (IE_red n t1)).
  revert Heq1.
  pattern (IE_red n t1) at 3 4.
  destruct (IE_red n t1); intros Heq1 Heq;
    try solve [simpl; eauto using SmallSteps_trans2, SmallSteps_app1, SmallSteps_app2].
  eapply SmallSteps_beta_iotafix.
    apply IHn.
    apply IHn.
    rewrite Heq; simpl.
    rewrite <- IE_subst_correction.
    apply IHn.
    apply IE_is_value_correction; auto.
  eapply SmallSteps_beta_iotafix.
    apply IHn.
    apply IHn.
    rewrite Heq; simpl.
    rewrite <- IE_subst_correction.
    change (TFix (IE_interp i)) with (IE_interp (IE_TFix i)).
    rewrite <- IE_subst_correction.
    apply IHn.
    apply IE_is_value_correction; auto.
  case_eq (IE_is_value (IE_red n t)); intro Heq0;
    [|simpl; eauto using SmallSteps_trans2, SmallSteps_match; fail].
  generalize (refl_equal (IE_red n t)).
  revert Heq0.
  pattern (IE_red n t) at 3 4.
  destruct (IE_red n t); intros Heq0 Heq;
    try solve [simpl; eauto using SmallSteps_match].
  simpl.
  assert (Hval : IsValue (IE_interp (IE_TConstr n0 l0))).
    rewrite Heq in Heq0.
    assert (Hval' := IE_is_value_correction Heq0).
    assumption.
  assert (Hclos : clos_after 0 (IE_interp (IE_TConstr n0 l0)))
    by (rewrite <- Heq; apply IE_value_is_closed; assumption).
  case_eq (nth_error l n0); [ intros (pn, pt) | ]; intro Heq_nth;
    [|eapply SmallSteps_match; eapply IHn; fail].
  case_eq (eq_nat_dec pn (length l0)); intros Heq_pn _;
    [|eapply SmallSteps_match; eapply IHn; fail].
  eapply SmallSteps_iota.
    eapply SmallSteps_trans2; [eapply IHn; fail|].
    rewrite Heq; simpl; reflexivity.
    match goal with
      |- nth_error (map ?f ?l) ?n = _ =>
        rewrite (map_nth_error f) with (d := (pn, pt))
    end; [|auto; fail].
    rewrite Heq_pn; rewrite map_length; reflexivity.
    rewrite IE_subst_list_iter_correction; rewrite map_rev; reflexivity.
    inversion Hval; auto; fail.
    simpl in Hclos.
    inversion Hclos; subst.
    clear - H1; rename H1 into Hclos.
    induction (map IE_interp l0); simpl in *; constructor.
      rewrite <- clos_alt; inversion Hclos; assumption.
      apply IHl; inversion Hclos; assumption.
  destruct e as [Hred Hxx'' Hcx']; simpl; assumption.
Qed.

(************************************************************************)

(** * Tactics *)

(** Applies to (simul t t'), where t has a product type. Transforms to goal
    into something like forall x x', ie_pred x x' -> simul (t x) (t' @ x') .
*)
Ltac ie_hnf :=
  lazymatch goal with
    | |- @simul (_ -> _) _ _ _ =>
      do 2 esplit;
        [ match goal with
            | |- ?t ===> _ =>
              IE_quote t ltac:(fun x =>
              let H := fresh in
              assert (H := IE_red_correction 10 x);
              simpl in H;
              eexact H)
            | |- ?g => fail 1 "ie_hnf: unable to prove" g
          end
        | split;
          [ match goal with
              | |- IsValue ?t =>
                IE_quote t ltac:(fun x =>
                apply (@IE_is_value_correction x);
                reflexivity)
              | |- ?g => fail 1 "ie_hnf: unable to prove" g
            end
          | match goal with
              | |- clos_after ?n ?t =>
                IE_quote t ltac:(fun x =>
                apply (@IE_is_closed_after_correction n x);
                reflexivity)
              | |- ?g => fail 1 "ie_hnf: unable to prove" g
            end
          | (* this is the resulting subgoal *)
          ]
        | match goal with
            | |- clos_after ?n ?t =>
              IE_quote t ltac:(fun x =>
              apply (@IE_is_closed_after_correction n x);
              reflexivity)
            | |- ?g => fail 1 "ie_hnf: unable to prove" g
          end
        ]; instantiate
    | _ => fail "ie_hnf: not a product"
  end
.

(** Applies to (simul t t'). Reduces t' as much as possible. *)
Ltac ie_simpl :=
  match goal with
    |- simul _ ?t' =>
      IE_quote t' ltac:(fun x =>
      fromtac;
        [ (* we simplify t w.r.t. SmallSteps reduction *)
          let H := fresh in
          assert (H := IE_red_correction 10 x);
          simpl in H;
            eexact H
        | (* we prove that t is closed *)
          apply (@IE_is_closed_after_correction _ x);
          reflexivity
        | (* this is the resulting subgoal *)
        ]; instantiate)
  end
.

(** Applies to [simul (f x) (f' @ x')]. Generates two subgoals [simul
   f f'] and [simul x x'] . *)
Ltac ie_apply :=
  (* typeclasses get lost somehow here *)
  lazymatch goal with |- simul (?a ?b) (?c @ ?d) =>
    refine (
      let H1 : simul a c := _ in
        let H2 : simul b d := _ in
          _ H1 H2)
  end; [ | |
    let H1 := fresh "Hie" in
    let H2 := fresh "Hie" in
      intros H1 H2; exact (ie_compat_app H1 H2)].

 (************************************************************************)
