(************************************************************************)
(*  v      * Internal Extraction                                        *)
(* <O___,, *                                                            *)
(*   \VV/  **************************************************************)
(*    //   * © 2007-2009 Stéphane Glondu                                *)
(*         * File distributed under the terms of the CeCILL 2.0 license *)
(************************************************************************)

Require Import Arith Max Omega Bool Setoid.
Require Import List.
Require Export MlSem.

Set Implicit Arguments.
Unset Strict Implicit.

(************************************************************************)

(** * Alternative definitions of closed terms *)

Inductive clos_after : nat -> term -> Prop :=
  | clos_after_TDummy :
    forall n,
      clos_after n TDummy
  | clos_after_TVar :
    forall n i,
      i < n -> clos_after n (TVar i)
  | clos_after_TFun :
    forall n t,
      clos_after (S n) t -> clos_after n (TFun t)
  | clos_after_TFix :
    forall n t,
      clos_after (S (S n)) t -> clos_after n (TFix t)
  | clos_after_TLet :
    forall n t1 t2,
      clos_after n t1 -> clos_after (S n) t2 -> clos_after n (TLet t1 t2)
  | clos_after_TApply :
    forall n t1 t2,
      clos_after n t1 -> clos_after n t2 -> clos_after n (TApply t1 t2)
  | clos_after_TConstr :
    forall n k tl,
      AllS (clos_after n) tl -> clos_after n (TConstr k tl)
  | clos_after_TMatch :
    forall n t pl,
      clos_after n t -> AllS (clos_after_pat n) pl -> clos_after n (TMatch t pl)
with clos_after_pat : nat -> pat -> Prop :=
  | clos_after_Patc :
    forall n di t,
      clos_after (di+n) t -> clos_after_pat n (Patc di t)
.

Hint Constructors clos_after.

Fixpoint is_clos_after n t :=
  match t with
    | TDummy => true
    | TVar i => if le_lt_dec n i then false else true
    | TFun t => is_clos_after (S n) t
    | TFix t => is_clos_after (S (S n)) t
    | TLet t1 t2 => is_clos_after n t1 && is_clos_after (S n) t2
    | TApply t1 t2 => is_clos_after n t1 && is_clos_after n t2
    | TConstr k tl => forallb (is_clos_after n) tl
    | TMatch t pl => is_clos_after n t && forallb (is_clos_after_pat n) pl
  end
with is_clos_after_pat n p :=
  match p with
    | Patc di t => is_clos_after (di+n) t
  end
.

(************************************************************************)

(** * Compatibility lemmas *)

Lemma is_clos_after_alt1 : forall n t, clos_after n t -> is_clos_after n t = true
  with is_clos_after_pat_alt1 : forall n p, clos_after_pat n p -> is_clos_after_pat n p = true.
Proof.
  (*  clos_after_alt1 *)
  intros n t H; destruct H as [ | | | | | | ? ? ? H | ? ? ? ? H ]; simpl;
    try solve [try rewrite andb_true_iff; intuition].
    destruct le_lt_dec; [ assert False by omega; contradiction | reflexivity ].
    induction H; simpl in *.
      reflexivity.
      rewrite andb_true_iff; intuition; fail.
    rewrite andb_true_iff; intuition.
    induction H; simpl in *.
      reflexivity.
      rewrite andb_true_iff; intuition; fail.
  (* clos_after_pat_alt1 *)
  destruct 1; simpl; auto; fail.
Qed.

Lemma is_clos_after_alt2 : forall n t, is_clos_after n t = true -> clos_after n t
  with is_clos_after_pat_alt2 : forall n p, is_clos_after_pat n p = true -> clos_after_pat n p.
Proof.
  Ltac bintuition := rewrite andb_true_iff; intuition; fail.
  (* clos_after_alt2 *)
  intros n t H; destruct t as [ | | | | | | ? tl | ? pl ]; simpl in *; constructor;
    try solve [try rewrite andb_true_iff in H; intuition].
    destruct le_lt_dec; [ discriminate | assumption ].
    induction tl; simpl in *; constructor; rewrite andb_true_iff in H; intuition; fail.
    rewrite andb_true_iff in H; destruct H as [_ H].
    induction pl; simpl in *; constructor; rewrite andb_true_iff in H; intuition; fail.
  (* clos_after_pat_alt2 *)
  intros n p H; destruct p; simpl; constructor; auto; fail.
Qed.

Lemma is_clos_after_alt :
  forall n t,
    clos_after n t <-> is_clos_after n t = true.
Proof.
  intuition (auto using is_clos_after_alt1, is_clos_after_alt2).
Qed.

Lemma is_clos_after_pat_alt :
  forall n p,
    clos_after_pat n p <-> is_clos_after_pat n p = true.
Proof.
  intuition (auto using is_clos_after_pat_alt1, is_clos_after_pat_alt2).
Qed.

Lemma clos_after'_alt1 : forall n t, clos_after' n t -> clos_after n t.
Proof.
  fix 2; intros n t H; destruct t as [ | | | | | | ? tl | ? pl];
    simpl in *; constructor;
    try solve [repeat rewrite clos_after_alt in H; intuition]; [ | ].
  induction tl; simpl in *; constructor.
    apply clos_after'_alt1; rewrite <- clos_after_alt; apply H; left; reflexivity.
    apply IHtl; intros u Hu; apply H; right; assumption.
    destruct H as [_ H]; induction pl as [ | [di t2] ? ]; simpl in *; constructor.
      match goal with |- _ _ ?X => specialize (H X) end; simpl in H.
      econstructor; apply clos_after'_alt1; rewrite <- clos_after_alt; eapply H; left; reflexivity.
      apply IHpl; intros p Hp; specialize (H p); destruct p; apply H; right; assumption.
Qed.

Lemma clos_after'_alt2 : forall n t, clos_after n t -> clos_after' n t.
Proof.
  fix 3; intros n t H;
    destruct H as [ | | | | | | ? ? ? H | ? ? ? ? H]; simpl in *;
    try solve [repeat rewrite clos_after_alt; intuition]; [ | ].
  induction H; intros u Hu; simpl in *.
    contradiction.
    destruct Hu as [Hu|Hu]; [subst u; rewrite clos_after_alt | ]; auto; fail.
      split.
      rewrite clos_after_alt; auto; fail.
  induction H; intros p Hp; simpl in *.
    contradiction.
    destruct Hp as [Hp|Hp].
      destruct H; subst; rewrite clos_after_alt; apply clos_after'_alt2; assumption.
      apply IHForall; assumption.
Qed.

Lemma clos_after'_alt :
  forall n t,
    clos_after' n t <-> clos_after n t.
Proof.
  intuition (auto using clos_after'_alt1, clos_after'_alt2).
Qed.

Lemma clos_alt : forall t, clos_after 0 t <-> clos t.
Proof.
  intros t; split.
  intros H n;
    rewrite <- clos_after'_alt, <- clos_after_alt in H;
    hnf in H; auto with arith; fail.
  intros H;
    rewrite <- clos_after'_alt, <- clos_after_alt;
    hnf in *; auto with arith; fail.
Qed.

(************************************************************************)