(************************************************************************)
(*  v      *   The Coq Proof Assistant  /  The Coq Development Team     *)
(* <O___,, * CNRS-Ecole Polytechnique-INRIA Futurs-Universite Paris Sud *)
(*   \VV/  **************************************************************)
(*    //   *      This file is distributed under the terms of the       *)
(*         *       GNU Lesser General Public License Version 2.1        *)
(************************************************************************)

(* Authors: Stéphane Glondu, Pierre Letouzey *)

Require Import Ml Arith List.

(** Substitution n with u in t (u is supposed closed).
    Variables (de Bruijn indexes) > n in t are decremented. *)

Fixpoint subst n u t {struct t} :=
  match t with
    | TDummy => TDummy
    | TVar k =>
        match lt_eq_lt_dec k n with
          | inleft (left _) => TVar k
          | inleft (right _) => u
          | inright _ => TVar (pred k)
        end
    | TFun t => TFun (subst (S n) u t)
    | TFix t => TFix (subst (S (S n)) u t)
    | TLet t1 t2 => TLet (subst n u t1) (subst (S n) u t2)
    | t1@t2 => (subst n u t1)@(subst n u t2)
    | TConstr k tl => TConstr k (map (subst n u) tl)
    | TMatch t pl =>
        TMatch (subst n u t)
          (map (fun p:pat => let (k,t):=p in Patc k (subst (k+n) u t)) pl)
  end
.

Notation "t [ n := u ]" := (subst n u t)
 (at level 8, n at next level, left associativity).

(** Multiple substitution:
    If l = [u0;u1;...], it replaces n by u0, n+1 by u1, etc.
    Variables > n+(length l) are decremented by (length l). *)

Fixpoint subst_list n l t {struct t} := match t with
   | TDummy => TDummy
   | TVar k => if (le_lt_dec n k) then nth (k-n) l (TVar (k-length l))
               else TVar k
   | TFun t => TFun (subst_list (S n) l t)
   | TFix t => TFix (subst_list (S (S n)) l t)
   | t1@t2 => (subst_list n l t1)@(subst_list n l t2)
   | TLet t1 t2 => TLet (subst_list n l t1) (subst_list (S n) l t2)
   | TConstr k tl => TConstr k (map (subst_list n l) tl)
   | TMatch t pl => TMatch (subst_list n l t)
         (map (fun p:pat => let (k,t):=p in Patc k (subst_list (k+n) l t)) pl)
  end
.

Notation "t [ n ::= l ]" := (subst_list n l t)
 (at level 8, n at next level, left associativity).

(** Alternative multiple substitution, by iterating subst *)

Fixpoint subst_list_iter n l t {struct l} := match l with
   | nil => t
   | u::l => subst_list_iter n l (t[n:=u])
  end
.

Notation "t [ n ;;= l ]" := (subst_list_iter n l t)
 (at level 8, n at next level, left associativity).
